(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import FMapIntMap.
Require Import Nnat.
Require Import Fomega.
Require Import Derived_multiple.
Require Import Derived_tupT.


(*************************************************************************)
(* Generic Rows *)
(*************************************************************************)

Definition lab := nat.

Inductive row (A : Set) : Set :=
  | nilR : row A
  | consR : lab -> A -> row A -> row A
.
Implicit Arguments nilR [A].
Implicit Arguments consR [A].

Fixpoint lengthR A (r : row A) {struct r} : nat :=
  match r with
  | nilR => 0
  | consR l a r' => 1 + lengthR A r'
  end.
Implicit Arguments lengthR [A].

Fixpoint getR A l (r : row A) {struct r} : option A :=
  match r with
  | nilR => None
  | consR l' a r' => if l === l' then Some a else getR A l r'
  end.
Implicit Arguments getR [A].

Definition field A l a (r : row A) : Prop := getR l r = Some a.
Implicit Arguments field [A].

Fixpoint labsR A (r : row A) {struct r} : list lab :=
  match r with
  | nilR => nil
  | consR l a r' => l :: labsR A r'
  end.
Implicit Arguments labsR [A].

Fixpoint fieldsR A (r : row A) {struct r} : list A :=
  match r with
  | nilR => nil
  | consR l a r' => a :: fieldsR A r'
  end.
Implicit Arguments fieldsR [A].

Fixpoint mapR A B f (r : row A) {struct r} : row B :=
  match r with
  | nilR => nilR
  | consR l a r' => consR l (f a) (mapR A B f r')
  end.
Implicit Arguments mapR [A B].

Fixpoint mapiR A B f (r : row A) {struct r} : row B :=
  match r with
  | nilR => nilR
  | consR l a r' => consR l (f l a) (mapiR A B f r')
  end.
Implicit Arguments mapiR [A B].

Fixpoint foldR A B f (b : B) (r : row A) {struct r} : B :=
  match r with
  | nilR => b
  | consR l a r' => f l a (foldR A B f b r')
  end.
Implicit Arguments foldR [A B].

(*
Definition to_mapR A (r : row A) : MapIntMap.t A :=
  foldR (fun l a M => MapIntMap.add (N_of_nat l) a M) (MapIntMap.empty A) r.
Definition of_mapR A (M : MapIntMap.t A) : row A :=
  MapIntMap.fold (fun n a r => consR (nat_of_N n) a r) M nilR.
Implicit Arguments to_mapR [A].
Implicit Arguments of_mapR [A].

Definition sortR A r : row A := of_mapR (to_mapR r).
Implicit Arguments sortR [A].
*)

Fixpoint catR A r1 r2 {struct r1} : row A :=
  match r1 with
  | nilR => r2
  | consR l a r1' => consR l a (catR A r1' r2)
  end.
(*
Definition catR A r1 r2 : row A :=
  of_mapR (MapIntMap.fold (fun n a M => MapIntMap.add n a M) (to_mapR r2) (to_mapR r1)).
*)
Implicit Arguments catR [A].

Fixpoint diffR A (r1 r2 : row A) {struct r1} :=
  match r1 with
  | nilR => nilR
  | consR l a r1' =>
    match getR l r2 with
    | None => consR l a (diffR A r1' r2)
    | Some _ => diffR A r1' r2
    end
  end.
(*
Definition diffR A (r1 r2 : row A) :=
  of_mapR (MapIntMap.fold (fun n a M => MapIntMap.remove n M) (to_mapR r2) (to_mapR r1)).
*)
Implicit Arguments diffR [A].

Lemma mapR_id : forall (A : Set) r,
  mapR (fun (x : A) => x) r = r.
Proof.
  intro A.
  induction r; simpl mapR.
    reflexivity.
    rewrite IHr. reflexivity.
Qed.

Lemma mapR_eta : forall (A : Set) (f : A -> A) r,
  mapR (fun x => f x) r = mapR f r.
Proof.
  intros A f. unfold mapR.
  induction r.
    reflexivity.
    rewrite IHr. reflexivity.
Qed.

Lemma mapR_mapR : forall (A : Set) (f : A -> A) (g : A -> A) r,
  mapR f (mapR g r) = mapR (fun x => f (g x)) r.
Proof.
  intros A. unfold mapR.
  induction r.
    reflexivity.
    rewrite IHr. reflexivity.
 Qed.

Lemma fieldsR_mapR : forall (A B : Set) (f : A -> B) r,
  fieldsR (mapR f r) = map f (fieldsR r).
Proof.
  intros A B f.
  induction r; simpl mapR; simpl fieldsR; simpl map.
    reflexivity.
    f_equal. apply IHr.
Qed.

Lemma fieldsR_length : forall A (r : row A),
  lengthR r = length (fieldsR r).
Proof.
  intro A.
  induction r; simpl lengthR; simpl fieldsR; simpl length.
    reflexivity.
    f_equal. apply IHr.
Qed.

Lemma mapR_length : forall (A B : Set) (f : A -> B) r,
  lengthR (mapR f r) = lengthR r.
Proof.
  intros A B f.
  induction r; simpl lengthR; simpl fieldsR; simpl length.
    reflexivity.
    f_equal. apply IHr.
Qed.

Lemma mapiR_length : forall (A B : Set) (f : lab -> A -> B) r,
  lengthR (mapiR f r) = lengthR r.
Proof.
  intros A B f.
  induction r; simpl lengthR; simpl fieldsR; simpl length.
    reflexivity.
    f_equal. apply IHr.
Qed.

Lemma catR_cons_right : forall (A : Set) l (a : A) r2 r1,
  exists r1',
  catR r1 (consR l a r2) = catR r1' r2.
Proof.
  intros A l a r2.
  induction r1; simpl catR.
    exists (consR l a nilR). simpl catR. reflexivity.
    destruct IHr1 as [r1' Eq]. rewrite Eq. exists (consR l0 a0 r1'). simpl catR. reflexivity.
Qed.

Lemma getR_cat_none : forall A l (r1 r2 : row A),
  getR l r1 = None ->
  getR l r2 = None ->
  getR l (catR r1 r2) = None.
Proof.
  intros A l.
  induction r1; intros r2 Hr1 Hr2; simpl catR; simpl getR in *.
    apply Hr2.
    destruct (l === l0); subst.
      apply Hr1.
      apply IHr1. apply Hr1. apply Hr2.
Qed.

Lemma getR_diff : forall A l (r2 r1 : row A),
  In l (labsR (diffR r1 r2)) ->
  getR l (diffR r1 r2) = getR l r1 /\
  getR l r2 = None.
Proof.
  intros A l r2.
  induction r1; intros Elem.
    inversion Elem.
    destruct (l === l0); subst.
      simpl diffR in *. destruct (getR l0 r2).
        destruct (IHr1 Elem) as [_ Eq]. inversion Eq.
        simpl getR. simpl labsR in Elem. simpl In in Elem. destruct Elem.
          destruct (l0 === l0). split; reflexivity. destruct (n H).
          destruct (IHr1 H) as [Eq _]. rewrite Eq. split; reflexivity.
      simpl diffR in *. destruct (getR l0 r2).
        simpl getR. destruct (l === l0). destruct (n e). apply (IHr1 Elem).
        simpl getR. simpl labsR in Elem. simpl In in Elem. destruct Elem.
          symmetry in H. destruct (n H).
          destruct (IHr1 H) as [Eq H2]. rewrite Eq. split. reflexivity. apply H2.
Qed.

Lemma getR_diff_none : forall A l (r1 r2 : row A),
  getR l r1 = None ->
  getR l (diffR r1 r2) = None.
Proof.
  intros A l.
  induction r1; intros r2 Hr1; simpl diffR; simpl getR in *.
    reflexivity.
    destruct (l === l0); subst; destruct (getR l0 r2).
      inversion Hr1.
      inversion Hr1.
      apply IHr1. apply Hr1.
      simpl getR. destruct (l === l0). destruct (n e). apply IHr1. apply Hr1.
Qed.

Lemma diffR_elem : forall A l (r2 r1 :row A),
  ~ In l (labsR r1) ->
  ~ In l (labsR (diffR r1 r2)).
Proof.
  intros A l r2.
  induction r1; intro H; simpl diffR; simpl labsR in *.
    apply H.
    destruct (l === l0); subst; destruct (getR l0 r2).
      destruct H. apply in_eq.
      destruct H. apply in_eq.
      apply IHr1. intro. destruct H. simpl In. right. apply H0.
      simpl labsR. simpl In. intro. destruct (l === l0).
        destruct (n e).
        destruct H0; subst.
          destruct n; reflexivity.
          apply IHr1. simpl in H. intro. destruct H. right. apply H1. apply H0.
Qed.


(*************************************************************************)
(* Specific Rows *)
(*************************************************************************)

Notation trow := (row typ).
Notation erow := (row exp).

Definition substATR a t := mapR (substAT a t).
Definition substAER a t := mapR (substAE a t).
Definition substXER x e := mapR (substXE x e).

Definition openATR := substATR 0.
Definition openAER := substAER 0.
Definition openXER := substXER 0.

Fixpoint substsATR (as' : list occ) (ts : list typ) rt {struct as'} : trow :=
  match as' with
  | nil => rt
  | a::as'' =>
    match ts with
    | nil => rt
    | t::ts' => substATR a t (substsATR as'' ts' rt)
    end
  end.

Fixpoint substsAER (as' : list occ) (ts : list typ) re {struct as'} : erow :=
  match as' with
  | nil => re
  | a::as'' =>
    match ts with
    | nil => re
    | t::ts' => substAER a t (substsAER as'' ts' re)
    end
  end.

Fixpoint substsXER (xs : list occ) (es : list exp) re {struct xs} : erow :=
  match xs with
  | nil => re
  | x::xs' =>
    match es with
    | nil => re
    | e::es' => substXER x e (substsXER xs' es' re)
    end
  end.

Fixpoint opensATR (ts : list typ) rt {struct ts} : trow :=
  match ts with
  | nil => rt
  | t'::ts' => substATR (length ts') t' (opensATR ts' rt)
  end.

Fixpoint opensAER (ts : list typ) re {struct ts} : erow :=
  match ts with
  | nil => re
  | t'::ts' => substAER (length ts') t' (opensAER ts' re)
  end.

Fixpoint opensXER (es : list exp) re {struct es} : erow :=
  match es with
  | nil => re
  | e'::es' => substXER (length es') e' (opensXER es' re)
  end.

(*************************************************************************)
(* Records *)
(*************************************************************************)

(* {l1:t1,...,lN:tN}   :=   t1 * ... * tN *)
Definition strT (r : row typ) := tupT (fieldsR r).

(* {l1=e1,...,lN=eN}   :=   (e1,...,eN) *)
Definition strE (r : row exp) := tupE (fieldsR r).

(* e.lI as {l1:t1,...,lN:tN}   :=   e.I *)
Fixpoint dotE' ls l {struct ls} : nat :=
  match ls with
  | nil => 0
  | l'::ls' => if l === l' then 1 else 1 + dotE' ls' l
  end.
Definition dotE (r : row typ) l e := projE (dotE' (labsR r) l) e.


(*************************************************************************)
(* Substitution *)
(*************************************************************************)

Lemma substATR_nil : forall a t',
  substATR a t' nilR = nilR.
Proof.
  intros a t'. unfold substATR. simpl mapR. reflexivity.
Qed.

Corollary openATR_nil : forall t',
  openATR t' nilR = nilR.
Proof.
  intro t'. unfold openATR. apply substATR_nil.
Qed.

Lemma substsATR_nil : forall as' ts,
  substsATR as' ts nilR = nilR.
Proof.
  induction as'; intro ts; simpl substsATR.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite IHas'. apply substATR_nil.
Qed.

Lemma opensATR_nil : forall ts,
  opensATR ts nilR = nilR.
Proof.
  induction ts; simpl opensATR.
    reflexivity.
    rewrite IHts. rewrite substATR_nil. reflexivity.
Qed.

Lemma substATR_cons : forall a t' l t rt,
  substATR a t' (consR l t rt) = consR l (substAT a t' t) (substATR a t' rt).
Proof.
  intros a t' l t rt. unfold substATR. simpl mapR. reflexivity.
Qed.

Corollary openATR_cons : forall t' l t rt,
  openATR t' (consR l t rt) = consR l (openAT t' t) (openATR t' rt).
Proof.
  intros t' l t rt. unfold openATR, openAT. apply substATR_cons.
Qed.

Lemma substsATR_cons : forall as' ts l t rt,
  substsATR as' ts (consR l t rt) = consR l (substsAT as' ts t) (substsATR as' ts rt).
Proof.
  induction as'; intros ts l t rt; simpl substsAT; simpl substsATR.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite IHas'. rewrite substATR_cons. reflexivity.
Qed.

Lemma opensATR_cons : forall ts l t rt,
  opensATR ts (consR l t rt) = consR l (opensAT ts t) (opensATR ts rt).
Proof.
  intros ts l t rt.
  induction ts; simpl opensAT; simpl opensATR.
    reflexivity.
    rewrite IHts. rewrite substATR_cons. reflexivity.
Qed.

Lemma substATR_cat : forall a t' rt1 rt2,
  substATR a t' (catR rt1 rt2) = catR (substATR a t' rt1) (substATR a t' rt2).
Proof.
  intros a t'.
  induction rt1; intro rt2; simpl catR; simpl substATR.
    reflexivity.
    rewrite substATR_cons. rewrite IHrt1. reflexivity.
Qed.

Corollary openATR_cat : forall t' rt1 rt2,
  openATR t' (catR rt1 rt2) = catR (openATR t' rt1) (openATR t' rt2).
Proof.
  intros t' rt1 rt2. unfold openATR. apply substATR_cat.
Qed.

Lemma substsATR_cat : forall as' ts rt1 rt2,
  substsATR as' ts (catR rt1 rt2) = catR (substsATR as' ts rt1) (substsATR as' ts rt2).
Proof.
  induction as'; intros ts rt1 rt2; simpl substsATR.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite IHas'. rewrite substATR_cat. reflexivity.
Qed.

Lemma opensATR_cat : forall ts rt1 rt2,
  opensATR ts (catR rt1 rt2) = catR (opensATR ts rt1) (opensATR ts rt2).
Proof.
  intros ts rt1 rt2.
  induction ts; simpl opensATR.
    reflexivity.
    rewrite IHts. rewrite substATR_cat. reflexivity.
Qed.


Lemma substXER_nil : forall x e',
  substXER x e' nilR = nilR.
Proof.
  intros x e'. unfold substXER. simpl mapR. reflexivity.
Qed.

Corollary openXER_nil : forall e',
  openXER e' nilR = nilR.
Proof.
  intro e'. unfold openXER. apply substXER_nil.
Qed.

Lemma substXER_cons : forall x e' l e re,
  substXER x e' (consR l e re) = consR l (substXE x e' e) (substXER x e' re).
Proof.
  intros x e' l e re. unfold substXER. simpl mapR. reflexivity.
Qed.

Corollary openXER_cons : forall e' l e re,
  openXER e' (consR l e re) = consR l (openXE e' e) (openXER e' re).
Proof.
  intros e' l e re. unfold openXER, openXE. apply substXER_cons.
Qed.


Lemma unfold_opensATR : forall ts rt,
  opensATR ts rt = substsATR (bnds (length ts)) ts rt.
Proof.
  induction ts; intro rt; simpl; try rewrite IHts; reflexivity.
Qed.

Lemma opensATR_concat : forall ts1 ts2 rt,
  opensATR (ts1 ++ ts2) rt =
  substsATR (map (shifts (length ts2)) (bnds (length ts1))) ts1 (opensATR ts2 rt).
Proof.
  induction ts1; intros ts2 rt; simpl.
    reflexivity.
    rewrite IHts1. rewrite app_length. rewrite shifts_bnd. do 2 f_equal. omega.
Qed.

Lemma substATR_substATR : forall t1 t2 rt a1 a2,
  a1 <> a2 ->
  indepAT a1 t2 ->
  indepAT a2 t1 ->
  substATR a1 t1 (substATR a2 t2 rt) = substATR a2 t2 (substATR a1 t1 rt).
Proof.
  intros t1 t2.
  induction rt; intros a1' a2' Neq H1 H2.
    do 2 rewrite substATR_nil. reflexivity.
    do 4 rewrite substATR_cons.
      rewrite IHrt. rewrite substAT_substAT.
        reflexivity.
        apply Neq. apply H1. apply H2.
        apply Neq. apply H1. apply H2.
Qed.

Lemma substATR_substsATR : forall as' ts a t rt,
  ~ In a as' ->
  indepATs a ts ->
  indepsAT as' t ->
  substATR a t (substsATR as' ts rt) = substsATR as' ts (substATR a t rt).
Proof.
  induction as'; intros ts a' t rt Ha Has' Hts; simpl substsATR.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite <- IHas'.
        rewrite substATR_substATR.
          reflexivity.
          intro Elem. apply Ha. rewrite Elem. apply in_eq.
          eapply indepATs_head. apply Has'.
          eapply indepsAT_head. apply Hts.
          intro Elem. apply Ha. apply in_cons. apply Elem.
          eapply indepATs_tail. apply Has'.
          eapply indepsAT_tail. apply Hts.
Qed.

Lemma substATR_opensATR : forall i t ts rt,
  length ts <= i ->
  lclosedT t ->
  lclosedTs ts ->
  substATR i t (opensATR ts rt) = opensATR ts (substATR i t rt).
Proof.
  intros i t.
  induction ts; intros rt Hi Ht Hts; simpl opensATR.
    reflexivity.
    rewrite <- IHts. rewrite substATR_substATR. reflexivity.
      simpl length in Hi. intros Elem. inversion Elem. omega.
      simpl indepAT. eapply lclosedTs_head. apply Hts.
      apply Ht.
      simpl length in Hi. omega.
      apply Ht.
      eapply lclosedTs_tail. apply Hts.
Qed.

Lemma substsATR_opensATR : forall is ts1 ts2 rt,
  (forall i, In i is -> length ts2 <= i) ->
  lclosedTs ts1 ->
  lclosedTs ts2 ->
  substsATR (map bnd is) ts1 (opensATR ts2 rt) = opensATR ts2 (substsATR (map bnd is) ts1 rt).
Proof.
  induction is; intros ts1 ts2 rt His Hts1 Hts2; simpl substsATR.
    reflexivity.
    destruct ts1.
      reflexivity.
      rewrite IHis.
        rewrite substATR_opensATR.
          reflexivity.
          apply His. apply in_eq.
          eapply lclosedTs_head. apply Hts1.
          apply Hts2.
          intros i Elem. apply His. apply in_cons. apply Elem.
          eapply lclosedTs_tail. apply Hts1.
          apply Hts2.
Qed.

(*
Lemma substsATR_opensATR : forall is ts1 ts2 rt,
  (forall i, In i is -> length ts <= i) ->
  lclosedTs ts1 ->
  lclosedTs ts2 ->
  substsATR (map bnd is) ts1 (opensATR ts2 rt) = opensATR ts2 (substsATR (map bnd is) ts1 rt).
Proof.
  induction as1; intros ts1 ts2 rt Has1 Hts2; simpl substsATR.
    reflexivity.
    destruct ts1.
      reflexivity.
      rewrite IHas1.
        rewrite substATR_opensATR.
          reflexivity.
          eapply lclosedTs_head. apply Hts1.
          apply Hts2.
          intros t' Elem. simpl indepAT. apply Has. apply in_eq. apply Elem.
          eapply lclosedTs_tail. apply Hts1.
          apply Hts2.
          intros a1 t2 Elem1 Elem2. apply Has.
            apply in_cons. apply Elem1.
            apply Elem2.
Qed.
*)

Lemma field_substAT : forall l a t' t rt,
  field l t rt ->
  field l (substAT a t' t) (substATR a t' rt).
Proof.
  intros l a t' t.
  induction rt; intros H. inversion H.
  unfold substATR. simpl mapR. fold (substATR a t' rt).
  unfold field in *. simpl getR in *.
  destruct (l === l0).
    inversion H. reflexivity.
    apply IHrt. apply H.
Qed.

Lemma field_openAT : forall l t' t rt,
  field l t rt ->
  field l (openAT t' t) (openATR t' rt).
Proof.
  intros l t' t rt H. unfold openAT, openATR. apply field_substAT. apply H.
Qed.

Lemma field_substsAT : forall l as' ts' t rt,
  field l t rt ->
  field l (substsAT as' ts' t) (substsATR as' ts' rt).
Proof.
  intros l.
  induction as'; intros ts' t rt H; simpl substsAT; simpl substsATR.
    apply H.
    destruct ts'.
      apply H.
      apply field_substAT. apply IHas'. apply H.
Qed.

Lemma field_opensAT : forall l ts' t rt,
  field l t rt ->
  field l (opensAT ts' t) (opensATR ts' rt).
Proof.
  intros l ts' t rt H. rewrite unfold_opensAT. rewrite unfold_opensATR. apply field_substsAT. apply H.
Qed.


Lemma fieldsR_substAT : forall a t' rt,
  fieldsR (substATR a t' rt) = map (substAT a t') (fieldsR rt).
Proof.
  intros a t' rt.
  induction rt; simpl fieldsR; simpl map.
    reflexivity.
    rewrite <- IHrt. reflexivity.
Qed.


Lemma mapR_opensAT : forall ts rt,
  mapR (opensAT ts) rt = opensATR ts rt.
Proof.
  intro t.
  induction rt; simpl mapR.
    rewrite opensATR_nil. reflexivity.
    rewrite opensATR_cons. rewrite IHrt. reflexivity.
Qed.

Lemma substAT_str : forall a t' rt,
  substAT a t' (strT rt) = strT (substATR a t' rt).
Proof.
  intros a t' rt.
  unfold strT.
  destruct rt.
    Case "nil".
      reflexivity.
    Case "cons".
      rewrite substAT_tup.
      rewrite fieldsR_substAT.
      reflexivity.
Qed.

Corollary openAT_str : forall t' rt,
  openAT t' (strT rt) = strT (openATR t' rt).
Proof.
  intros t' rt. unfold openAT, openATR. apply substAT_str.
Qed.

Lemma substsAT_str : forall as' ts rt,
  substsAT as' ts (strT rt) = strT (substsATR as' ts rt).
Proof.
  induction as'; intros ts rt; simpl substsAT; simpl substsATR.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite IHas'. rewrite substAT_str. reflexivity.
Qed.

Lemma opensAT_str : forall ts rt,
  opensAT ts (strT rt) = strT (opensATR ts rt).
Proof.
  intros ts rt.
  induction ts; simpl opensAT; simpl opensATR.
    reflexivity.
    rewrite IHts. rewrite substAT_str. reflexivity.
Qed.


Lemma fieldsR_substAE : forall a t' re,
  fieldsR (substAER a t' re) = map (substAE a t') (fieldsR re).
Proof.
  intros a t' re.
  induction re; simpl fieldsR; simpl map.
    reflexivity.
    rewrite <- IHre. reflexivity.
Qed.

Lemma substAE_str : forall a t re,
  substAE a t (strE re) = strE (substAER a t re).
Proof.
  intros a t re.
  unfold strE.
  destruct re.
    Case "nil".
      reflexivity.
    Case "cons".
      rewrite substAE_tup.
      rewrite fieldsR_substAE.
      reflexivity.
Qed.

Corollary openAE_str : forall t re,
  openAE t (strE re) = strE (openAER t re).
Proof.
  intros t re. unfold openAE, openAER. apply substAE_str.
Qed.


Lemma fieldsR_substXE : forall x e' re,
  fieldsR (substXER x e' re) = map (substXE x e') (fieldsR re).
Proof.
  intros x e' re.
  induction re; simpl fieldsR; simpl map.
    reflexivity.
    rewrite <- IHre. reflexivity.
Qed.

Lemma substXE_str : forall x e' re,
  substXE x e' (strE re) = strE (substXER x e' re).
Proof.
  intros x e' re.
  unfold strE.
  destruct re.
    Case "nil".
      reflexivity.
    Case "cons".
      rewrite substXE_tup.
      rewrite fieldsR_substXE.
      reflexivity.
Qed.

Corollary openXE_str : forall e' re,
  openXE e' (strE re) = strE (openXER e' re).
Proof.
  intros e' re. unfold openXE, openXER. apply substXE_str.
Qed.



Lemma labsR_substATR : forall a t' rt,
  labsR (substATR a t' rt) = labsR rt.
Proof.
  intros a t' rt.
  induction rt; simpl labsR; simpl map.
    reflexivity.
    rewrite <- IHrt. reflexivity.
Qed.

Lemma labsR_opensATR : forall ts rt,
  labsR (opensATR ts rt) = labsR rt.
Proof.
  intros ts.
  induction rt.
    rewrite opensATR_nil. reflexivity.
    rewrite opensATR_cons. simpl labsR. f_equal. apply IHrt.
Qed.


Lemma substAE_dot : forall a t rt l e,
  substAE a t (dotE rt l e) = dotE (substATR a t rt) l (substAE a t e).
Proof.
  intros a t rt l e.
  unfold dotE.
  rewrite substAE_proj.
  rewrite labsR_substATR.
  reflexivity.
Qed.

Corollary openAE_dot : forall t rt l e,
  openAE t (dotE rt l e) = dotE (openATR t rt) l (openAE t e).
Proof.
  intros t rt l e. unfold openAE, openATR. apply substAE_dot.
Qed.

Lemma substsAE_dot : forall as' ts rt l e,
  substsAE as' ts (dotE rt l e) = dotE (substsATR as' ts rt) l (substsAE as' ts e).
Proof.
  induction as'; intros ts rt l e; simpl substsAE; simpl substsATR.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite IHas'. rewrite substAE_dot. reflexivity.
Qed.

Lemma opensAE_dot : forall ts rt l e,
  opensAE ts (dotE rt l e) = dotE (opensATR ts rt) l (opensAE ts e).
Proof.
  induction ts; intros rt l e; simpl opensAE; simpl opensATR.
    reflexivity.
    rewrite IHts. rewrite substAE_dot. reflexivity.
Qed.

Lemma substXE_dot : forall x e' rt l e,
  substXE x e' (dotE rt l e) = dotE rt l (substXE x e' e).
Proof.
  intros x e' rt l e.
  unfold dotE.
  rewrite substXE_proj.
  reflexivity.
Qed.

Corollary openXE_dot : forall e' rt l e,
  openXE e' (dotE rt l e) = dotE rt l (openXE e' e).
Proof.
  intros e' rt l e. unfold openXE. apply substXE_dot.
Qed.


(*************************************************************************)
(* Local closure *)
(*************************************************************************)

Inductive lclosedTR : trow -> Prop :=
  | lclosedTR_nil :
      lclosedTR nilR
  | lclosedTR_cons : forall l t rt,
      lclosedT t ->
      lclosedTR rt ->
      lclosedTR (consR l t rt)
.

Lemma lclosedTs_fieldsR : forall rt,
  lclosedTR rt ->
  lclosedTs (fieldsR rt).
Proof.
  induction rt; intros Hrt; simpl fieldsR.
    apply lclosedTs_nil.
    inversion Hrt; subst. apply lclosedTs_cons. apply H1. apply IHrt. apply H3.
Qed.

Lemma lclosedT_str : forall rt,
  lclosedTR rt ->
  lclosedT (strT rt).
Proof.
  intros rt Hrt.
  unfold strT.
  apply lclosedT_tup.
  apply lclosedTs_fieldsR. apply Hrt.
Qed.


Inductive lclosedER : erow -> Prop :=
  | lclosedER_nil :
      lclosedER nilR
  | lclosedER_cons : forall l e re,
      lclosedE e ->
      lclosedER re ->
      lclosedER (consR l e re)
.

Lemma lclosedEs_fieldsR : forall re,
  lclosedER re ->
  lclosedEs (fieldsR re).
Proof.
  induction re; intros Hre; simpl fieldsR.
    apply lclosedEs_nil.
    inversion Hre; subst. apply lclosedEs_cons. apply H1. apply IHre. apply H3.
Qed.

Lemma lclosedE_str : forall re,
  lclosedER re ->
  lclosedE (strE re).
Proof.
  intros re Hre.
  unfold strE.
  apply lclosedE_tup.
  apply lclosedEs_fieldsR. apply Hre.
Qed.


Lemma lclosedE_dot : forall rt l e,
  lclosedTR rt ->
  lclosedE e ->
  lclosedE (dotE rt l e).
Proof.
  intros rt l e Hrt He.
  unfold dotE.
  apply lclosedE_proj.
  apply He.
Qed.


(*************************************************************************)
(* Typing *)
(*************************************************************************)

Inductive ofTR : tenv -> row typ -> Prop :=
  | ofTR_nil : forall D,
      okD D ->
      ofTR D nilR
  | ofTR_cons : forall D l t rt,
      ofT D t omK ->
      ofTR D rt ->
      ofTR D (consR l t rt)
.

Lemma validityDTR : forall D rt,
  ofTR D rt ->
  okD D.
Proof.
  intro D.
  induction rt; intro Hrt; inversion Hrt; subst.
    apply H.
    apply IHrt. apply H4.
Qed.

Lemma ofTR_cat : forall D rt1 rt2,
  ofTR D rt1 ->
  ofTR D rt2 ->
  ofTR D (catR rt1 rt2).
Proof.
  intro D.
  induction rt1; intros rt2 Hrt1 Hrt2; simpl catR.
    apply Hrt2.
    inversion Hrt1; subst. apply ofTR_cons.
      apply H2.
      apply IHrt1. apply H4. apply Hrt2.
Qed.

 
Lemma ofTR_diff : forall D rt1 rt2,
  ofTR D rt1 ->
  ofTR D (diffR rt1 rt2).
Proof.
  intro D.
  induction rt1; intros rt2 Hrt1; simpl diffR.
    apply Hrt1.
    inversion Hrt1; subst. destruct (getR l rt2).
      apply IHrt1. apply H4.
      apply ofTR_cons. apply H2. apply IHrt1. apply H4.
Qed.

Lemma ofTs_fieldsR : forall D rt,
  ofTR D rt ->
  (forall t, In t (fieldsR rt) -> ofT D t omK).
Proof.
  intro D.
  induction rt; intros Hrt t Elem; simpl fieldsR in *.
    inversion Elem.
    inversion Hrt; subst. inversion Elem; subst.
      apply H2.
      apply IHrt. apply H4. apply H.
Qed.

Lemma ofT_str : forall D rt,
  ofTR D rt ->
  ofT D (strT rt) omK.
Proof.
  intros D rt Hrt.
  unfold strT.
  apply ofT_tup.
    eapply validityDTR. apply Hrt.
    apply ofTs_fieldsR. apply Hrt.
Qed.

Lemma ofT_str_inversion : forall D rt,
  ofT D (strT rt) omK ->
  ofTR D rt.
Proof.
  intros D rt Ht.
  unfold strT in Ht.
  induction rt; simpl tupT in Ht.
    apply ofTR_nil. eapply validityDT. apply Ht.
    inversion Ht; subst. apply ofTR_cons.
      apply H2.
      apply IHrt. apply H3.
Qed.


Inductive ofER : tenv -> eenv -> row exp -> row typ -> Prop :=
  | ofER_nil : forall D G,
      okG D G ->
      ofER D G nilR nilR
  | ofER_cons : forall D G l e t re rt,
      ofE D G e t ->
      ofER D G re rt ->
      ofER D G (consR l e re) (consR l t rt)
.

Lemma ofER_cat : forall D G re1 re2 rt1 rt2,
  ofER D G re1 rt1 ->
  ofER D G re2 rt2 ->
  ofER D G (catR re1 re2) (catR rt1 rt2).
Proof.
  intros D G.
  induction re1; intros re2 rt1 rt2 Hre1 Hre2; destruct rt1; simpl catR.
    apply Hre2.
    inversion Hre1.
    inversion Hre1.
    inversion Hre1; subst. apply ofER_cons.
      apply H3.
      apply IHre1. apply H8. apply Hre2.
Qed.

Lemma validityGER : forall D G re rt,
  ofER D G re rt ->
  okG D G.
Proof.
  intros D G.
  induction re; intros rt Hre; inversion Hre; subst.
    apply H.
    eapply IHre. apply H6.
Qed.

Corollary validityDER : forall D G re rt,
  ofER D G re rt ->
  okD D.
Proof.
  intros D G re rt Hre. eapply validityDG. eapply validityGER. apply Hre.
Qed.

Lemma ofEs_fieldsR : forall D G re rt,
  ofER D G re rt ->
  (forall e t, In (e, t) (combine (fieldsR re) (fieldsR rt)) -> ofE D G e t).
Proof.
  intros D G.
  induction re; intros rt Hre e t Elem; destruct rt; simpl fieldsR in *.
    inversion Elem.
    inversion Elem.
    inversion Hre; subst.
    inversion Hre; subst. simpl combine in Elem. inversion Elem.
      inversion H; subst. apply H3.
      eapply IHre. apply H8. apply H.
Qed.

Lemma length_ofER : forall D G re rt,
  ofER D G re rt ->
  length (fieldsR re) = length (fieldsR rt).
Proof.
  intros D G.
  induction re; intros rt Hre; destruct rt; simpl fieldsR in *; inversion Hre; subst.
    reflexivity.
    simpl length. f_equal. apply IHre. apply H8.
Qed.

Lemma ofE_str : forall D G re rt,
  ofER D G re rt ->
  ofE D G (strE re) (strT rt).
Proof.
  intros D G re rt Hre.
  unfold strE, strT.
  apply ofE_tup.
    eapply validityGER. apply Hre.
    eapply length_ofER. apply Hre.
    apply ofEs_fieldsR. apply Hre.
Qed.


Lemma getR_dotE' : forall l (t : typ) rt,
  getR l rt = Some t ->
  exists n,
  dotE' (labsR rt) l = S n.
Proof.
  intros l t.
  induction rt; simpl labsR; simpl getR; simpl dotE'; intro H. inversion H.
  destruct (l === l0); subst.
    exists 0. reflexivity.
    destruct (IHrt H) as [n' H']. exists (S n'). omega.
Qed.

Lemma getR_nth : forall l t rt n,
  getR l rt = Some t ->
  S n = dotE' (labsR rt) l ->
  t = nth n (fieldsR rt) unitT.
Proof.
  intros l t.
  induction rt; intros n Hl Hn; inversion Hl.
  simpl dotE' in Hn. simpl labsR in Hn. simpl fieldsR.
  destruct (l === l0); inversion Hn; inversion H0; subst.
    simpl nth. reflexivity.
    destruct (getR_dotE' l t rt H0) as [n H3].
      rewrite H3. simpl nth. apply IHrt. apply H0. symmetry. apply H3.
Qed.

Lemma dotE'_length : forall l t (rt : trow) n,
  getR l rt = Some t ->
  S n = dotE' (labsR rt) l ->
  n < length (fieldsR rt).
Proof.
  intros l t.
  induction rt; simpl labsR; simpl fieldsR; simpl getR; simpl length; intros n Hl H.
    inversion H.
    simpl dotE' in H. destruct (l === l0); subst; inversion H.
      omega.
      destruct (getR_dotE' l t rt Hl) as [n' H2]. rewrite H2.
      assert (H3 : n' < length (fieldsR rt)).
        apply IHrt. apply Hl. symmetry. apply H2.
      omega.
Qed.

Lemma ofE_dot : forall D G l t rt e,
  ofE D G e (strT rt) ->
  field l t rt ->
  ofE D G (dotE rt l e) t.
Proof.
  intros D G l t.
  unfold dotE. unfold strT.
  induction rt; intros e He Hl; simpl labsR; simpl dotE'; inversion Hl.
  destruct (l === l0); inversion H0; subst.
    simpl tupT in He. simpl projE. eapply ofE_fst. apply He.
    destruct (getR_dotE' l t rt H0) as [n' H2]. rewrite H2 in *.
      simpl projE in *. fold (projE (S n') (sndE e)).
      symmetry in H2. rewrite (getR_nth l t rt n' H0 H2).
      apply ofE_proj.
        split. omega. eapply dotE'_length. apply H0. apply H2.
        eapply ofE_snd. simpl fieldsR in He. simpl tupT in He. apply He.
Qed.


Fixpoint distinct (A : Set) (xs : list A) {struct xs} : Prop :=
  match xs with
  | nil => True
  | x::xs' => ~ In x xs' /\ distinct A xs'
  end.
Implicit Arguments distinct [A].

Lemma distinct_diffR : forall A (r2 r1 : row A),
  distinct (labsR r1) ->
  distinct (labsR (diffR r1 r2)).
Proof.
  intros A r2.
  induction r1; simpl labsR; intro H.
    apply H.
    simpl in H. destruct H. destruct (getR l r2).
      apply (IHr1 H0).
      simpl labsR. simpl. split.
        apply diffR_elem. apply H.
        apply (IHr1 H0).
Qed.

Lemma ofER_dots' : forall C G e rt rt',
  distinct (labsR rt') ->
  (forall l, In l (labsR rt') -> getR l rt' = getR l rt) ->
  ofE C G e (strT rt) ->
  ofER C G (mapiR (fun l _ => dotE rt l e) rt') rt'.
Proof.
  intros C G e rt.
  induction rt'; intros Hrt1 Hrt2 He; simpl mapiR.
    apply ofER_nil. eapply validityGE. apply He.
    apply ofER_cons.
      apply ofE_dot. apply He. unfold field. rewrite <- Hrt2.
        simpl getR. destruct (l === l). reflexivity. destruct n; reflexivity.
        simpl labsR. apply in_eq.
        simpl fieldsR in Hrt1. simpl distinct in Hrt1. destruct Hrt1.
      apply IHrt'.
        apply H0.
        intros l' Elem. destruct (l === l'); subst.
          destruct (H Elem).
          rewrite <- Hrt2.
            simpl getR. destruct (l' === l); subst. destruct n; reflexivity. reflexivity.
            simpl labsR. apply in_cons. apply Elem.
       apply He.
Qed.

Corollary ofER_dots : forall C G e rt,
  distinct (labsR rt) ->
  ofE C G e (strT rt) ->
  ofER C G (mapiR (fun l _ => dotE rt l e) rt) rt.
Proof.
  intros C G e rt Hrt He.
  apply ofER_dots'.
    apply Hrt. intros; reflexivity. apply He.
Qed.


Lemma ofEs_dots' : forall D G e t x rt rt',
  distinct (labsR rt') ->
  (forall l, In l (labsR rt') -> getR l rt' = getR l rt) ->
  ofE D G x (strT rt) ->
  In (e, t) (combine (fieldsR (mapiR (fun l _ => dotE rt l x) rt')) (fieldsR rt')) ->
  ofE D G e t.
Proof.
  intros D G e t x rt.
  induction rt'; intros Dis Sub Hx Elem; simpl in Elem; inversion Elem.
    inversion H; subst. apply ofE_dot.
      apply Hx.
      unfold Derived_strT.field. rewrite <- Sub.
        simpl getR. destruct (l === l). reflexivity. destruct n; reflexivity.
        simpl labsR. apply in_eq.
    apply IHrt'.
      simpl in Dis. destruct Dis. apply H1.
      intros l' Elem'. rewrite <- Sub.
        simpl getR. destruct (l' === l); subst.
          simpl in Dis. destruct Dis. destruct (H0 Elem').
          reflexivity.
        simpl labsR. simpl in Dis. apply in_cons. apply Elem'.
      apply Hx.
      apply H.
Qed.

Corollary ofEs_dots : forall D G e t x rt,
  distinct (labsR rt) ->
  ofE D G x (strT rt) ->
  In (e, t) (combine (fieldsR (mapiR (fun l _ => dotE rt l x) rt)) (fieldsR rt)) ->
  ofE D G e t.
Proof.
  intros D G e t x rt Dis Hx Elem.
  eapply ofEs_dots'. apply Dis. intros l Elem'; reflexivity. apply Hx. apply Elem.
Qed.


(*Not needed so far.
Lemma eqT_str : forall rt rt',
  labsR rt = labsR rt' ->
  (forall t t', In (t, t') (combine (fieldsR rt) (fieldsR rt')) -> eqT t t') ->
  eqT (strT rt) (strT rt').
Proof.
*)


(*************************************************************************)
(* Weakening for rows *)
(*************************************************************************)

Lemma weakeningDTR : forall D1 D D2 rt,
  ofTR (D1 & D2) rt ->
  okD (D1 & D & D2) ->
  ofTR (D1 & D & D2) rt.
Proof.
  intros D1 D D2 rt. intros Hrt.
  remember (D1 & D2) as D12.
  generalize dependent D2.
  induction Hrt; intros D2 Eq HD'; subst.
    Case "nil".
      apply ofTR_nil.
      apply HD'.
    Case "cons".
      apply ofTR_cons.
        apply weakeningDT. apply H. apply HD'.
        apply IHHrt. reflexivity. apply HD'.
Qed.

Lemma weakeningDTR_head : forall D1 D2 rt,
  ofTR D1 rt ->
  okD (D1 & D2) ->
  ofTR (D1 & D2) rt.
Proof.
  intros D1 D2 rt. intros Hrt HD'.
  replace (D1 & D2) with (D1 & D2 & empty).
    apply weakeningDTR.
      rewrite concat_empty. apply Hrt.
      rewrite concat_empty. apply HD'.
    rewrite concat_empty. reflexivity.
Qed.

Lemma weakeningDER : forall D1 D D2 G re rt,
  ofER (D1 & D2) G re rt ->
  okD (D1 & D & D2) ->
  ofER (D1 & D & D2) G re rt.
Proof.
  intros D1 D D2 G re rt. intros Hre.
  remember (D1 & D2) as D'.
  generalize dependent D2.
  induction Hre; intros D2 Eq HD'; subst.
    Case "nil".
      apply ofER_nil.
      apply weakeningDG. apply H. apply HD'.
    Case "cons".
      apply ofER_cons.
        apply weakeningDE. apply H. apply HD'.
        apply IHHre. reflexivity. apply HD'.
Qed.

Lemma weakeningDER_head : forall D1 D2 G re rt,
  ofER D1 G re rt ->
  okD (D1 & D2) ->
  ofER (D1 & D2) G re rt.
Proof.
  intros D1 D2 G re rt. intros Hre HD.
  replace (D1 & D2) with (D1 & D2 & empty).
    apply weakeningDER.
      rewrite concat_empty. apply Hre.
      rewrite concat_empty. apply HD.
    rewrite concat_empty. reflexivity.
Qed.

Lemma weakeningGER : forall D G1 G G2 re rt,
  ofER D (G1 & G2) re rt ->
  okG D (G1 & G & G2) ->
  ofER D (G1 & G & G2) re rt.
Proof.
  intros D G1 G G2 re rt. intros Hre.
  remember (G1 & G2) as G'.
  generalize dependent G2.
  induction Hre; intros G2 Eq HG'; subst.
    Case "nil".
      apply ofER_nil.
      apply HG'.
    Case "cons".
      apply ofER_cons.
        apply weakeningGE. apply H. apply HG'.
        apply IHHre. reflexivity. apply HG'.
Qed.

Lemma weakeningGER_head : forall D G1 G2 re rt,
  ofER D G1 re rt ->
  okG D (G1 & G2) ->
  ofER D (G1 & G2) re rt.
Proof.
  intros D G1 G2 re rt. intros Hre HG.
  replace (G1 & G2) with (G1 & G2 & empty).
    eapply weakeningGER.
      rewrite concat_empty. apply Hre.
      rewrite concat_empty. apply HG.
    rewrite concat_empty. reflexivity.
Qed.
