(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import Metatheory_Misc.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived.
Require Import Modules_syntax.
Require Import Modules_semantic.
Require Import Modules_elaboration.
Require Import Modules_matching.


(*************************************************************************)
(* Matching *)
(*************************************************************************)

Corollary sound_matchS : forall C s s' ks ts e,
  matchS C s s' ks ts e ->
  forall L,
  ofT C  (transS s) omK ->
  (forall bs, fresh L (length ks) bs -> ofT (C & bs ~~ ks)  (opensAT (freeTs bs) (transS s')) omK )->
  ofE C empty e (arrT (transS s) (opensAT ts (transS s'))).
Proof.
  eapply proj1. apply sound_enrich.
Qed.

 
(*************************************************************************)
(* Facts about translated environments *)
(*************************************************************************)

Lemma transG_fresh : forall C G X s x,
  wfG C G ->
  binds' X s x G ->
  x \in dom (transG G).
Proof.
  intros C G X s x HG HX.
  induction G; simpl transG.
    inversion HX.
    destruct a as [[X' s'] x']. env_fix. simpl dom.
      unfold binds' in HX. destruct (x == x'); subst.
        apply S.union_2. apply S.singleton_2. apply S.E.eq_refl.
        apply S.union_3. apply IHG.
          inversion HG. apply H3.
          simpl get' in HX. destruct (X === X').
            inversion HX. symmetry in H1. destruct (n H1).
            apply HX.
Qed.

Lemma trans_binds' : forall C G X s x,
  wfG C G ->
  binds' X s x G ->
  binds x (transS s) (transG G).
Proof.
  intros C G X s x HG HX.
  induction G; simpl transG.
    inversion HX.
    destruct a as [[X' s'] x']. env_fix.
      unfold binds' in HX. simpl get' in HX. destruct (X === X').
        inversion HX; subst. apply binds_head.
        inversion HG; subst. apply binds_tail.
          intro Eq. subst. absurd (x' \in dom (transG G)).
            apply H5.
            eapply transG_fresh with (X := X) (s := s). apply H3. apply HX.
          apply IHG. apply H3. apply HX.
Qed.


(*************************************************************************)
(* Environment validity *)
(*************************************************************************)

Lemma validityGG : forall G C,
  wfG C G ->
  okG C (transG G).
Proof.
  intros G C HG.
  induction HG.
    Case "empty".
      simpl. apply okG_empty. apply H.
    Case "var".
      replace (transG (G &' X~'s@x)) with (transG G & x~transS s).
        2: simpl; reflexivity.
      apply okG_var.
        apply IHHG.
        apply H.
        apply H0.
Qed.

Lemma validityGG_inv : forall G C,
  okG C (transG G) ->
  wfG C G.
Proof.
  intros G C HG.
  induction G.
    Case "nil".
      apply wfG_empty. eapply validityDG. apply HG.
    Case "cons".
      destruct a as [[X s] x].
      inversion HG. subst.
      assert (Aux : forall A (x : A) xs, x :: xs = (x :: nil) ++ xs). auto.
      rewrite Aux.
      apply wfG_var.
        apply IHG. apply H3.
        apply H4.
(*        apply lclosedS_trans. eapply regularityT. apply H5.*)
        apply H5.
Qed.

Lemma validityCG : forall C G,
  wfG C G ->
  okD C.
Proof.
  intros C G HG.
  induction G; inversion HG.
    apply H.
    apply IHG. apply H1.
Qed.


(*************************************************************************)
(* Soundness of elaboration *)
(*************************************************************************)

Theorem sound_elab :
  (forall C G T k t,
   C; G |=T T ~: k ~> t ->
   okG C (transG G) /\ ofT C t k) /\
  (forall C G E t e,
   C; G |=E E ~: t ~> e ->
   ofE C (transG G) e t) /\
  (forall C G S q,
   C; G |=S S ~> q ->
   okG C (transG G) /\ ofT C (transQ q) omK) /\
  (forall C G D q,
   C; G |=D D ~> q ->
   okG C (transG G) /\ ofT C (transQ q) omK) /\
  (forall C G M q e,
   C; G |=M M ~: q ~> e ->
   ofE C (transG G) e (transQ q)) /\
  (forall C G B q e,
   C; G |=B B ~: q ~> e ->
   ofE C (transG G) e (transQ q)) /\
  (forall C G P s e,
   C; G |=P P ~: s ~> e ->
   ofE C (transG G) e (transS s)).
Proof.
  apply IIelab.
    Case "TypT".
      intros C G t k HG Ht.
      split. apply validityGG. apply HG.
      apply Ht.
    Case "PathT".
      intros C G P t k e HP IHe.
      simpl transS in IHe.
      assert (Ht' : ofT C (typT t k) omK). eapply validityTE. apply IHe.
      split. eapply validityGE. apply IHe.
      eapply ofT_typ_inversion.
        eapply lclosedT_typ_inversion. eapply regularityT. apply Ht'.
        apply Ht'.
    Case "PackT".
      intros C G S q HS IHGq.
      apply IHGq.

    Case "ExpE".
      intros C G e t He.
      apply He.
    Case "PathE".
      intros C G P t e HP IHe.
      simpl transS in IHe.
      apply IHe.
    Case "PackE".
      intros C G X S s' x ks s ts f HX HS [IHG IHs] Hts.
      inversion Hts; subst.
      simpl transQ in *.
      apply ofE_packs.
        symmetry. apply H.
        apply H0.
        apply ofE_app with (t2 := transS s').
          replace (transG G) with (empty & transG G). 2: rewrite empty_concat; reflexivity.
            apply weakeningGE_head.
              assert (HG : wfG C G).
               apply validityGG_inv; apply IHG.
              assert (binds_x := trans_binds' C G X s' x HG HX).
              assert (s'Ok : ofT C (transS s') omK).
               eapply validityTG. apply validityGG. apply HG.
               apply binds_x.
              assert (IHs_inv:=ofT_exs_inversion _ _ _ IHs).
              destruct IHs_inv as [L K]. (* we need L now,  before applying sound_matchS below *)
              eapply sound_matchS. apply Hts.
               apply s'Ok.
               apply K.
              rewrite <- empty_concat. apply IHG.
          apply ofE_var.
            apply IHG.
            eapply trans_binds'. apply validityGG_inv. apply IHG. apply HX.
        apply IHs.

    Case "PathS".
      intros C G P q e HP IHe.
      simpl transS in IHe.
      split. eapply validityGE. apply IHe.
      assert (Hq' : ofT C (typT (transQ q) omK) omK). eapply validityTE. apply IHe.
      eapply ofT_typ_inversion.
        eapply lclosedT_typ_inversion. eapply regularityT. apply Hq'.
        apply Hq'.
    Case "StrS".
      intros C G D q HD [IHG IHq].
      split. apply IHG.
      apply IHq.
    Case "FunS".
      intros L C G X S1 S2 ks s q HS1 [IHG IHs] HS2 IHG'q.
      destruct (ofT_exs_inversion _ _ _ IHs) as [L1 IHs'].
      simpl transQ.
      split. apply IHG.
      apply ofT_alls with (L := L \u L1). eapply validityDT. apply IHs.
        intros as1 Frs1.
        rewrite opensAT_arr.
        apply ofT_arr.
          simpl transQ in IHs. apply IHs'. fresh_solve.
          rewrite opensAT_transQ. 2: apply lclosedTs_vars.
            pick_fresh_gen L x.
            eapply proj2. apply IHG'q with (x := x). fresh_solve. apply Fr.
    Case "TypEqS".
      intros L C G S Xs ks1 k ks2 s s' T t HS [IHG IHq] HT [_ IHt] Hs.
      simpl transQ in *.
      destruct (ofT_exs_inversion _ _ _ IHq) as [L1 IHs].
      split. apply IHG.
      apply ofT_exs with (L := L \u L1 \u dom C). eapply validityDT. apply IHt.
      intros as' Frs.
      rewrite app_length in Frs.
      destruct (fresh_partition _ ks1 ks2 L as') as [as1 [as2 [Eq [Len1 Len2]]]].
        fresh_solve.
      pick_fresh_gen (L \u L1 \u varsAs as1 \u varsAs as2) a.
      assert (Len2' : length (k::ks2) = length (a::as2)). simpl. f_equal. omega.
      rewrite app_length in IHs.
      rewrite <- Eq in *. rewrite Len1 in *. rewrite Len2 in *. rewrite Len2' in *.
      assert (Frs' : fresh (L \u L1) (length as1 + length (a::as2)) (as1 ++ a::as2)).
        apply fresh_concat_l.
          destruct (fresh_concat_r _ _ _ Frs). fresh_solve.
          simpl. split.
            notin_solve.
            apply fresh_union_l.
              destruct (fresh_concat_r _ _ _ Frs). fresh_solve.
              apply fresh_single.
                notin_solve.
                destruct (fresh_concat_r _ _ _ Frs). fresh_solve.
      assert (HC1 : okD (C & as1~~ks1)).
        apply okD_vars. eapply validityDT. apply IHt.
           destruct (fresh_concat_r _ _ _ Frs). rewrite Len1. fresh_solve.
      assert (HC2 : okD (C & as1~~ks1 & as2~~ks2)).
        apply okD_vars. apply HC1.
           destruct (fresh_concat_r _ _ _ Frs). rewrite Len2.
             rewrite dom_concat. rewrite dom_multi. fresh_solve. apply Len1.
      rewrite opensAT_transS. 2: apply lclosedTs_vars.
      rewrite Hs with (a0 := a). 2: fresh_solve. 2: omega.
      rewrite <- substAT_transS. 2: eapply regularityT; apply IHt.
      rewrite <- opensAT_transS. 2: apply lclosedTs_vars.
      rewrite concat_multi_app. 2: apply Len1. 2: apply Len2.
      rewrite <- concat_assoc.
      eapply substitutabilityAT.
        rewrite concat_assoc. rewrite <- concat_multi_cons.
          rewrite concat_assoc. rewrite <- concat_multi_app.
            2: apply Len1. 2: simpl; f_equal; apply Len2.
          apply IHs. fresh_solve.
        apply weakeningDT_head. apply weakeningDT_head.
          apply IHt. apply HC1. apply HC2.

    Case "ValD".
      intros C G X T t HT [IHG IHt].
      simpl transQ.
      split. apply IHG.
      apply ofT_str.
      apply ofTR_cons.
        apply IHt.
        apply ofTR_nil. eapply validityDT. apply IHt.
    Case "TypD".
      intros C G X K HG.
      simpl transQ.
      split. apply validityGG. apply HG.
      eapply ofT_ex. intros a Fr.
      assert (HC' : okD (C & a~K)).
        apply okD_var.
          eapply validityCG. apply HG.
          inversion HG. apply Fr. apply Fr.
      rewrite openAT_str.
      apply ofT_str.
      rewrite openATR_cons.
      apply ofTR_cons.
        unfold typT, openAT. simpl substAT.
          fold (typT a K).
          apply ofT_typ.
          apply ofT_var. apply HC'. apply binds_head.
        rewrite openATR_nil. apply ofTR_nil. apply HC'.
    Case "TypEqD".
      intros C G X T t k HT [IHG IHt].
      simpl transQ.
      split. apply IHG.
      apply ofT_str.
      apply ofTR_cons.
        apply ofT_typ. apply IHt.
        apply ofTR_nil. eapply validityDT. apply IHt.
    Case "ModD".
      intros C G X S ks s HS [IHG IHs].
      simpl transQ in *.
      split. apply IHG.
      destruct (ofT_exs_inversion _ _ _ IHs) as [L IHs'].
      apply ofT_exs with (L := L \u dom C).
        eapply validityDT. apply IHs.
        intros as' Fr.
          rewrite opensAT_str.
          apply ofT_str.
          rewrite opensATR_cons.
          apply ofTR_cons.
            apply IHs'. fresh_solve.
            rewrite opensATR_nil. apply ofTR_nil. apply okD_vars.
              eapply validityDT. apply IHs.
              fresh_solve.
    Case "SigEqD".
      intros C G X S q HS [IHG IHq].
      simpl transQ.
      split. apply IHG.
      apply ofT_str.
      apply ofTR_cons.
        apply ofT_typ. apply IHq.
        apply ofTR_nil. eapply validityDT. apply IHq.
    Case "EmpD".
      intros C G HG.
      simpl transQ.
      split. apply validityGG. apply HG.
      apply ofT_str.
      apply ofTR_nil.
      eapply validityCG. apply HG.
    Case "SeqD".
      intros L C G D1 D2 ks1 r1 r1' ks2 r2 Hr1 Hr2 HD1 [IHG IHr1] HD2 IHG'r2 Heq Hdiff.
      simpl transQ in *.
      split. apply IHG.
      rewrite exsT_concat.
      destruct (ofT_exs_inversion _ _ _ IHr1) as [L1 IHr1'].
      apply ofT_exs with (L := L \u L1 \u dom C).
        eapply validityDT. apply IHr1.
        intros as1 Fr1.
          assert (Fr1' : fresh L (length ks1) as1).
            eapply proj1. apply fresh_union_r.
            eapply proj1. apply fresh_union_r.
            apply Fr1.
          pick_freshes_gen L (sizeR r1) xs1.
          destruct (IHG'r2 as1 xs1 Fr1' Fr) as [_ IHr2].
          rewrite opensAQ_ex in IHr2.
          simpl transQ in IHr2.
          destruct (ofT_exs_inversion _ _ _ IHr2) as [L2 IHr2'].
          rewrite opensAT_exs.
          apply ofT_exs with (L := L \u varsAs as1 \u L1 \u L2 \u dom C).
            eapply validityDT. apply IHr2.
            intros as2 Fr2.
              assert (Len1 : length ks1 = length as1).
                eapply fresh_length. apply Fr1.
              assert (Len2 : length ks2 = length as2).
                eapply fresh_length. apply Fr2.
              rewrite substsAT_str. rewrite opensAT_str.
              apply ofT_str.
              rewrite transR_seq.
              rewrite substsATR_cat. rewrite opensATR_cat.
              rewrite substsATR_transR. 2: apply lclosedTs_vars.
              rewrite opensATR_transR. 2: apply lclosedTs_vars.
              apply ofTR_cat.
                (* left *)
                  assert (Fr11 : fresh L1 (length ks1) as1).
                    eapply proj2. apply fresh_union_r.
                    eapply proj1. apply fresh_union_r.
                    apply Fr1.
                  assert (IHr1'' := IHr1' as1 Fr11).
                  rewrite opensAT_str in IHr1''.
                  assert (IHr1''' := ofT_str_inversion _ _ IHr1'').
                  rewrite <- opensATR_transR. 2: apply lclosedTs_vars.
                  rewrite <- substsATR_transR. 2: apply lclosedTs_vars.
                  replace (length ks2) with (length (freeTs as2)).
                    2: rewrite freeTs_length. 2: symmetry. 2: eapply fresh_length. 2: apply Fr2.
                  rewrite map_shifts_bnds.
                  rewrite <- substsATR_opensATR.
                    rewrite <- map_shifts_bnds. rewrite <- opensATR_concat.
                      rewrite <- freeTs_concat.
                      rewrite opensATR_transR. 2: apply lclosedTs_vars.
                      rewrite Heq.
                        rewrite <- opensATR_transR. 2: apply lclosedTs_vars.
                          eapply weakeningDTR_head. apply IHr1'''.
                            eapply validityDT. apply IHr2'.
                              eapply proj2; apply fresh_union_r.
                              eapply proj1; apply fresh_union_r.
                              apply Fr2.
                        rewrite Len1. rewrite Len2. apply fresh_concat_l.
                          rewrite <- Len1. apply Fr1'.
                          rewrite <- Len2. do 3 (eapply proj1; apply fresh_union_r). apply Fr2.
                        symmetry. apply Len1.
                    intros i Elem.
                      assert (Idx : forall n m i, In i (map (fun j => n + j) (enum m)) -> n <= i).
                        intro n. induction m; intros i' Elem'; simpl enum in *; simpl map in *.
                          inversion Elem'.
                          destruct Elem'. omega. apply IHm. apply H.
                      eapply Idx. apply Elem.
                    apply lclosedTs_vars.
                    apply lclosedTs_vars.
                (* right *)
                  assert (Fr22 : fresh L2 (length ks2) as2).
                    eapply proj2. apply fresh_union_r.
                    eapply proj1. apply fresh_union_r.
                    apply Fr2.
                  assert (IHr2'' := IHr2' as2 Fr22).
                  rewrite <- substsAT_transS in IHr2''. 2: apply lclosedTs_vars.
                  simpl transS in IHr2''.
                  rewrite substsAT_str in IHr2''.
                  rewrite opensAT_str in IHr2''.
                  apply (ofT_str_inversion _ _ IHr2'').
    Case "InclD".
      intros C G S q HS [IHG IHq].
      split. apply IHG.
      apply IHq.

    Case "IdM".
      intros C G X s x HG HX.
      simpl transQ.
      apply ofE_var.
        apply validityGG. apply HG.
        eapply trans_binds'. apply HG. apply HX.
    Case "StrM".
      intros C G B q e HB IHe.
      apply IHe.
    Case "DotM".
      intros C G M X ks r s e e' Hr HM IHe HX He'.
      simpl transQ in *.
      assert (IHt : ofT C (exsT ks (strT (transR r))) omK).
        eapply validityTE. apply IHe.
      destruct (ofT_exs_inversion _ _ _ IHt) as [L IHt'].
      assert (IHs : ofT C (exsT ks (transS s)) omK).
        apply ofT_exs with (L := L).
          eapply validityDT. apply IHt.
          intros as' Fras.
            assert (IHt'' := IHt' as' Fras).
            assert (HX' : field X (opensAS (freeTs as') s) (opensAR (freeTs as') r)).
              apply field_opens. apply HX.
            rewrite opensAT_transS. 2: apply lclosedTs_vars.
            rewrite opensAT_str in IHt''.
            rewrite opensATR_transR in IHt''. 2: apply lclosedTs_vars.
            eapply ofT_field.
              apply ofT_str_inversion. apply IHt''.
              apply HX'.
      apply ofE_unpacks with (L := L \u dom C \u dom (transG G)).
        simpl transS. apply IHe.
        intros as' x Fras' Frx.
          assert (Len : length as' = length ks).
            symmetry. eapply fresh_length. apply Fras'.
          assert (HC' : okD (C & as' ~~ ks)).
            eapply validityDT. apply IHt'.
            do 2 (eapply proj1; apply fresh_union_r). apply Fras'.
          rewrite (He' as'). 2: symmetry; apply Len.
          rewrite openXE_packs. 2: rewrite freeTs_length. 2: apply Len.
          apply ofE_packs.
            rewrite freeTs_length. apply Len.
            intros t k Elem.
              destruct (t_in_as t k as' ks Elem) as [a Eq]. subst.
              apply ofT_vars.
                apply HC'.
                apply (proj1 (In_rev (combine as' ks) (pair a k))). apply in_freeTs. apply Elem.
            rewrite opensAE_dot. rewrite openXE_dot. apply ofE_dot.
              rewrite openXE_opensAE. 2: apply lclosedE_var.
                unfold openXE. simpl substXE.
                rewrite opensAE_lclosed. 2: apply lclosedE_var.
                apply ofE_var.
                  apply okG_var.
                    eapply weakeningDG_head. eapply validityGE. apply IHe. apply HC'.
                    notin_solve.
                    simpl transS. apply IHt'.
                      do 2 (eapply proj1; apply fresh_union_r). apply Fras'.
                  simpl transS. rewrite opensAT_str. apply binds_head.
              apply field_opensAT. apply field_trans. apply HX.
            eapply weakeningDT_head. apply IHs. apply HC'.
        apply IHs.
    Case "FunM".
      intros L C G X S M ks s q e HS [IHG IHs] HM IHe.
      simpl transQ in *.
      destruct (ofT_exs_inversion _ _ _ IHs) as [L' IHs'].
      apply ofE_gens with (L := L \u L').
        apply IHG.
        intros as' Fras.
          rewrite opensAE_lam. rewrite opensAT_arr.
          apply ofE_lam with (L := L \u L').
            apply IHs'. fresh_solve.
            intros x Frx.
              rewrite opensAT_transS. 2: apply lclosedTs_vars.
              replace (x ~ transS (opensAS (freeTs as') s)) with (transG (X ~' opensAS (freeTs as') s @ x)).
                2: simpl transG; reflexivity.
              rewrite <- transG_concat.
              rewrite opensAT_transQ. 2: apply lclosedTs_vars.
              apply IHe. fresh_solve. notin_solve.
    Case "AppM".
      intros C G X1 X2 ks s q x1 s2 x2 ts f HG HX1 HX2 Hts.
      inversion Hts; subst.
      apply ofE_app with (t2 := opensAT ts (transS s)).
        rewrite <- opensAT_transQ. rewrite <- opensAT_arr.
          apply ofE_insts with (ks := ks).
            apply ofE_var.
              apply validityGG. apply HG.
              replace (allsT ks (arrT (transS s) (transQ q))) with (transS (funS ks s q)).
                2: simpl transS; reflexivity.
                eapply trans_binds'. apply HG. apply HX1.
            apply H.
            apply H0.
          eapply regularityTs. apply H. apply H0.
        apply ofE_app with (t2 := transS s2).
          replace (transG G) with (empty & transG G). 2: rewrite empty_concat; reflexivity.
            apply weakeningGE_head.
              assert (binds_x1 := trans_binds' C G X1 (funS ks s q) x1 HG HX1).
              assert (funsigOk : ofT C (transS (funS ks s q)) omK).
               eapply validityTG. apply validityGG. apply HG.
               apply binds_x1.
               simpl transS in funsigOk.
               assert (IfunsigOk:=ofT_alls_inversion _ _ _ funsigOk).
              destruct IfunsigOk as [L K]. (* we need L now,  before applying sound_matchS below *)
              eapply sound_matchS. apply Hts.
               assert (binds_x2 := trans_binds' C G X2 s2 x2 HG HX2).
               eapply validityTG. apply validityGG. apply HG.
               apply binds_x2.
               intros bs frLbs.
               assert (bodyOk : ofT (C & bs ~~ ks) (opensAT (freeTs bs) (arrT (transS s) (transQ q)))omK ).
                apply K; apply frLbs. 
                rewrite opensAT_arr in bodyOk.
                inversion bodyOk; auto.
              rewrite <- empty_concat. apply validityGG. apply HG.
          apply ofE_var.
            apply validityGG. apply HG.
            eapply trans_binds'. apply HG. apply HX2.
    Case "SealM".
      intros C G X S s' x ks s ts f HX HS [IHG IHs] Hts.
      inversion Hts; subst.
      simpl transQ in *.
      apply ofE_packs.
        symmetry. apply H.
        apply H0.
        apply ofE_app with (t2 := transS s').
          replace (transG G) with (empty & transG G). 2: rewrite empty_concat; reflexivity.
            apply weakeningGE_head.
              assert (HG : wfG C G).
               apply validityGG_inv; apply IHG.
              assert (binds_x := trans_binds' C G X s' x HG HX).
              assert (s'Ok : ofT C (transS s') omK).
               eapply validityTG. apply validityGG. apply HG.
               apply binds_x.
              assert (IHs_inv:=ofT_exs_inversion _ _ _ IHs).
              destruct IHs_inv as [L K]. (* we need L now,  before applying sound_matchS below *)
              eapply sound_matchS. apply Hts.
               apply s'Ok.
               apply K.
              rewrite <- empty_concat. apply IHG.
          apply ofE_var.
            apply IHG.
            eapply trans_binds'. apply validityGG_inv. apply IHG. apply HX.
        apply IHs.
    Case "UnpackM".
      intros C G E S q e HE IHe HS IHq.
      apply IHe.

    Case "ValB".
      intros C G X E t e HE IHe.
      simpl transQ.
      apply ofE_str.
      apply ofER_cons.
        apply IHe.
        apply ofER_nil. eapply validityGE. apply IHe.
    Case "TypB".
      intros C G X T t k HT [IHG IHt].
      simpl transQ.
      apply ofE_str.
      apply ofER_cons.
        apply ofE_typ.
          apply IHG.
          apply IHt.
        apply ofER_nil.
          apply IHG.
    Case "ModB".
      intros C G X M ks s e e' HM IHe He'.
      simpl transQ in *. simpl transS.
      assert (IHt : ofT C (exsT ks (transS s)) omK). eapply validityTE. apply IHe.
      destruct (ofT_exs_inversion _ _ _ IHt) as [L IHs].
      apply ofE_unpacks with (L := L \u dom (transG G)).
        apply IHe.
        intros as' x Fras' Frx.
          assert (Len : length (freeTs as') = length ks).
            symmetry; rewrite freeTs_length; eapply fresh_length; apply Fras'.
          assert (Fras'' : fresh L (length ks) as').
            eapply proj1. apply fresh_union_r. apply Fras'.
          assert (HG' : okG (C & as' ~~ ks) (transG G & x ~ opensAT (freeTs as') (transS s))).
            apply okG_var.
              eapply weakeningDG_head.
                eapply validityGE. apply IHe.
                eapply validityDT. apply IHs. apply Fras''.
              notin_solve.
              apply IHs. apply Fras''.
          rewrite He'. 2: eapply fresh_length; apply Fras'.
          rewrite openXE_packs. 2: apply Len.
          apply ofE_packs.
            apply Len.
            intros t k Elem.
              destruct (t_in_as t k as' ks Elem) as [a Eq]. subst.
              apply ofT_vars.
                eapply validityDT. apply IHs. apply Fras''.
                apply (proj1 (In_rev (combine as' ks) (pair a k))). apply in_freeTs. apply Elem.
                rewrite openXE_str. rewrite opensAT_str. apply ofE_str.
                  rewrite openXER_cons. rewrite opensATR_cons. apply ofER_cons.
                    unfold openXE. simpl substXE. apply ofE_var.
                      apply HG'.
                      apply binds_head.
                    rewrite openXER_nil. rewrite opensATR_nil. apply ofER_nil. apply HG'.
            apply ofT_exs with (L := L \u dom (C & as' ~~ ks)).
              eapply validityDG. apply HG'.
              intros as'' Fras'''.
                assert (HD'' : okD (C & as' ~~ ks & as'' ~~ ks)).
                  apply okD_vars.
                    eapply validityDG. apply HG'.
                    eapply proj2. apply fresh_union_r. apply Fras'''.
                rewrite opensAT_str. apply ofT_str.
                  rewrite opensATR_cons. apply ofTR_cons.
                    eapply weakeningDT.
                      apply IHs. eapply proj1. apply fresh_union_r. apply Fras'''.
                      apply HD''.
                    rewrite opensATR_nil. apply ofTR_nil. apply HD''.
        apply ofT_exs with (L := L).
          eapply validityDT. apply IHt.
          intros as'' Fras''.
            rewrite opensAT_str. apply ofT_str.
            rewrite opensATR_cons. apply ofTR_cons.
              apply IHs. apply Fras''.
              rewrite opensATR_nil. apply ofTR_nil. eapply validityDT. apply IHs. apply Fras''.
    Case "SigB".
      intros C G X S q HS [IHG IHq].
      simpl transQ.
      apply ofE_str.
      apply ofER_cons.
        apply ofE_typ.
          apply IHG.
          apply IHq.
        apply ofER_nil.
          apply IHG.
    Case "EmpB".
      intros C G HG.
      simpl transQ.
      apply ofE_str.
      apply ofER_nil.
        apply validityGG. apply HG.
    Case "SeqB".
      intros L C G B1 B2 ks1 r1 r1' e1 ks2 r2 e2 es' e' Hr1 Hr2 HB1 IHe1 HB2 IHe2 Hr1' Hes' He'.
      simpl transQ in *.
      rewrite exsT_concat.
      assert (HC : okD C).
        eapply validityDE. apply IHe1.
      assert (IHt1 : ofT C (exsT ks1 (strT (transR r1))) omK).
        eapply validityTE. apply IHe1.
      destruct (ofT_exs_inversion _ _ _ IHt1) as [L1 IHt1'].
      assert (Ht : ofT C (exsT ks1 (exsT ks2 (strT (transR (seqR r1' r2))))) omK).
        apply ofT_exs with (L := L \u L1 \u dom C). apply HC.
        intros as1 Fras1.
        assert (HC1 : okD (C & as1~~ks1)).
          apply okD_vars. apply HC. fresh_solve.
        assert (Fras11 : fresh L1 (length ks1) as1). fresh_solve.
        assert (IHr1' := IHt1' as1 Fras11).
        rewrite opensAT_str in IHr1'.
        assert (IHr1 := ofT_str_inversion _ _ IHr1').
        assert (IHt2 : ofT (C & as1~~ks1) (opensAT (freeTs as1) (exsT ks2 (strT (transR r2)))) omK).
          pick_freshes_gen L (sizeR r1) xs.
          assert (IHe2' := IHe2 as1 xs).
          rewrite <- opensAT_transQ in IHe2'. 2: apply lclosedTs_vars. simpl transQ in IHe2'.
          eapply validityTE. apply IHe2'. fresh_solve. fresh_solve.
        rewrite opensAT_exs in IHt2.
        destruct (ofT_exs_inversion _ _ _ IHt2) as [L2 IHt2'].
        rewrite opensAT_exs.
        apply ofT_exs with (L := L \u L1 \u L2 \u dom C \u varsAs as1). apply HC1.
        intros as2 Fras2.
        assert (Len1 : length ks1 = length as1).
          eapply fresh_length. apply Fras1.
        assert (Len2 : length ks2 = length as2).
          eapply fresh_length. apply Fras2.
        assert (HC2 : okD (C & as1~~ks1 & as2~~ks2)).
          apply okD_vars. apply HC1. rewrite dom_concat. rewrite dom_multi. fresh_solve. omega.
        assert (Fras22 : fresh L2 (length ks2) as2). fresh_solve.
        assert (IHr2' := IHt2' as2 Fras22).
        rewrite substsAT_str in IHr2'. rewrite opensAT_str in IHr2'.
        assert (IHr2 := ofT_str_inversion _ _ IHr2').
        rewrite substsAT_str. rewrite opensAT_str.
        apply ofT_str.
        rewrite map_shifts_bnds.
        rewrite <- substsATR_opensATR.
          rewrite <- map_shifts_bnds. rewrite Len2. rewrite <- freeTs_length. rewrite <- opensATR_concat.
            rewrite opensATR_transR.
              rewrite opensAR_seq. rewrite transR_seq. apply ofTR_cat.
                rewrite <- freeTs_concat. rewrite Hr1'.
                  rewrite opensAR_diff. rewrite transR_diff. apply ofTR_diff.
                    rewrite opensATR_transR in IHr1. 2: apply lclosedTs_vars.
                      eapply weakeningDTR_head. apply IHr1. apply HC2.
                  rewrite Len1. rewrite Len2. apply fresh_concat_l.
                    rewrite <- Len1. fresh_solve.
                    rewrite <- Len2. fresh_solve.
                  omega.
              rewrite <- opensATR_transR.
                rewrite opensATR_concat. rewrite map_shifts_bnds.
                rewrite substsATR_opensATR.
                  rewrite <- map_shifts_bnds. rewrite freeTs_length. rewrite <- Len2. apply IHr2.
                  intros i Elem. eapply bnd_upper. apply Elem.
                  apply lclosedTs_vars.
                  apply lclosedTs_vars.
                apply lclosedTs_concat. apply lclosedTs_vars. apply lclosedTs_vars.
              apply lclosedTs_concat. apply lclosedTs_vars. apply lclosedTs_vars.
          intros i Elem.
            rewrite freeTs_length. rewrite <- Len2.
            clear IHt2 IHt2' IHr2 IHr2'.
            induction (length (freeTs as1)); simpl enum in Elem; simpl map in Elem; inversion Elem.
              omega.
              apply IHn. apply H.
          apply lclosedTs_vars.
          apply lclosedTs_vars.
      (* And now for the term... *)
      assert (HG : okG C (transG G)).
        eapply validityGE. apply IHe1.
      apply ofE_unpacks with (L := L \u L1 \u dom C \u dom (transG G)).
        simpl transS. apply IHe1. (* e1 *)
        intros as1 x1 Fras1 Frx1.
          assert (Len1 : length as1 = length ks1).
            symmetry. eapply fresh_length. apply Fras1.
          assert (HC1 : okD (C & as1~~ks1)).
            apply okD_vars. apply HC. fresh_solve.
          assert (HG' : okG (C & as1~~ks1) (transG G)).
            eapply weakeningDG_head. apply HG. apply HC1.
          assert (HG1 : okG (C & as1~~ks1) (transG G & x1 ~ strT (transR (opensAR (freeTs as1) r1)))).
            assert (Fras1L1 : fresh L1 (length ks1) as1). fresh_solve.
            assert (IHt1'' := IHt1' as1 Fras1L1). rewrite opensAT_str in IHt1''. rewrite opensATR_transR in IHt1''.
            apply okG_var. apply HG'. fresh_solve. apply IHt1''. apply lclosedTs_vars.
          assert (Fras1L : fresh L (length ks1) as1). fresh_solve.
          pick_freshes_gen L (sizeR r1) xs'.
          assert (IHe2' := IHe2 as1 xs' Fras1L Fr).
          rewrite <- opensAT_transQ in IHe2'. 2: apply lclosedTs_vars. simpl transQ in IHe2'.
          rewrite opensAT_exs in IHe2'. rewrite substsAT_str in IHe2'.
          rewrite substsATR_transR in IHe2'. 2: apply lclosedTs_vars.
          assert (IHt2' := validityTE _ _ _ _ IHe2').
          destruct (ofT_exs_inversion _ _ _ IHt2') as [L2 IHt2].
          rewrite opensAE_unpacks. rewrite openXE_unpacks.
          apply ofE_unpacks with (L := L \u L1 \u L2 \u dom C \u varsAs as1 \u dom (transG G) \u {{x1}}).
            rewrite opensAE_lets. rewrite openXE_lets. apply ofE_lets with (L := L \u {{x1}} \u dom (transG G)).
              rewrite map_length. rewrite Hes'.
                simpl. do 2 rewrite <- fieldsR_length. rewrite mapiR_length. do 2 rewrite transR_length. apply sizeR_opensAR.
                omega.
              intros ei ti Elem. (* x.l_i : t_i *)
                rewrite Hes' in Elem. 2: omega.
                simpl transS. rewrite opensAT_str.
                simpl in Elem. rewrite <- fieldsR_mapR in Elem.
                apply ofEs_dots with (x := x1) (rt := opensATR (freeTs as1) (transR r1)).
                  rewrite labsR_opensATR. rewrite <- labsR_transR. apply Hr1.
                  apply ofE_var.
                    rewrite opensATR_transR. apply HG1.
                    apply lclosedTs_vars.
                    apply binds_head.
                  rewrite opensATR_transR. rewrite mapR_opensAT in Elem. rewrite opensATR_transR in Elem. apply Elem.
                  apply lclosedTs_vars.
                  apply lclosedTs_vars.
              intros xs Frxs. (* e2 *)
                rewrite map_length in Frxs.
                assert (Lenxs : length (fieldsR (transR r1)) = length xs). eapply fresh_length. apply Frxs.
                rewrite <- substXE_opensXE.
                  2: rewrite freeEs_length; rewrite map_length; omega.
                  2: apply lclosedE_var.
                  2: apply lclosedEs_vars.
                assert (FrxsL : fresh L (sizeR r1) xs).
                  rewrite <- fieldsR_length in Frxs. rewrite transR_length in Frxs. fresh_solve.
                apply weakeningGE.
                  rewrite substXE_lclosed.
                    rewrite <- fieldsR_mapR. rewrite mapR_opensAT. rewrite opensATR_transR. 2: apply lclosedTs_vars. 
                    assert (IHe2'' := IHe2 as1 xs).
                    rewrite transG_concat in IHe2''. rewrite transG_str in IHe2''.
                    rewrite opensAQ_ex in IHe2''. simpl transQ in IHe2''.
                    rewrite substsAT_transS. 2: apply lclosedTs_vars.
                    apply IHe2''. fresh_solve. apply FrxsL.
                    eapply regularityE. apply IHe2. fresh_solve. apply FrxsL.
                 apply okG_vars.
                   simpl transS. rewrite opensAT_str. rewrite opensATR_transR. 2: apply lclosedTs_vars. apply HG1.
                   intros t Elem.
                     assert (Fras11 : fresh L1 (length ks1) as1). fresh_solve.
                     assert (IHr1' := IHt1' as1 Fras11).
                     rewrite opensAT_str in IHr1'.
                     assert (IHr1 := ofT_str_inversion _ _ IHr1').
                     rewrite <- fieldsR_mapR in Elem. rewrite mapR_opensAT in Elem.
                     clear IHr1'. generalize dependent Elem. generalize dependent IHr1.
                     induction (opensATR (freeTs as1) (transR r1)); intros IHr1 Elem; simpl fieldsR in Elem; inversion Elem; inversion IHr1; subst.
                       apply H3.
                       apply IHr. apply H5. apply H.
                   simpl dom. rewrite map_length. fresh_solve.
            intros as2 x2 Fras2 Frx2. (* pack *)
              assert (Len2 : length as2 = length ks2).
                symmetry. eapply fresh_length. apply Fras2.
              assert (HC2 : okD (C & as1~~ks1 & as2~~ks2)).
                apply okD_vars. apply HC1. rewrite dom_concat. rewrite dom_multi. fresh_solve. omega.
              assert (IHr2 : ofTR (C & as1~~ks1 & as2~~ks2) (transR (opensAR (freeTs (as1 ++ as2)) r2))).
                assert (Fras2L2 : fresh L2 (length ks2) as2). fresh_solve.
                assert (IHr2 := IHt2 as2 Fras2L2).
                rewrite opensAT_str in IHr2. rewrite <- substsATR_transR in IHr2. 2: apply lclosedTs_vars.
                rewrite <- Len2 in IHr2. rewrite <- freeTs_length in IHr2. rewrite map_shifts_bnds in IHr2.
                rewrite <- substsATR_opensATR in IHr2.
                  rewrite <- map_shifts_bnds in IHr2.
                    rewrite <- opensATR_concat in IHr2.
                    rewrite <- freeTs_concat in IHr2.
                    rewrite opensATR_transR in IHr2. 2: apply lclosedTs_vars.
                    apply ofT_str_inversion. apply IHr2.
                  intros i Elem. eapply bnd_upper. apply Elem.
                  apply lclosedTs_vars.
                  apply lclosedTs_vars.
              rewrite <- substXE_opensAE. 2: apply lclosedE_var.
              replace (length ks2) with (length (freeTs as2)). 2: rewrite freeTs_length; omega.
              rewrite <- opensAE_concat. 2: apply lclosedTs_vars. 2: apply lclosedTs_vars.
              rewrite <- exsT_concat. rewrite <- freeTs_concat.
              rewrite freeTs_length. rewrite Len2.
              rewrite He'. 2: omega. 2: omega.
              apply ofE_packs.
                rewrite freeTs_length. do 2 rewrite app_length. omega.
                intros t k Elem. (* ai : ki *)
                  rewrite concat_assoc. rewrite <- concat_multi_app. 2: omega. 2: omega.
                  apply ofT_var_tail.
                    do 2 rewrite app_length. omega.
                    rewrite concat_multi_app. 2: omega. 2: omega. rewrite <- concat_assoc. apply HC2.
                    apply Elem.
                rewrite opensAT_str. apply ofE_str. (* strE *)
                  rewrite opensATR_transR. 2: apply lclosedTs_vars.
                  rewrite opensAR_seq. rewrite transR_seq.
                  simpl transS. rewrite substsAT_str. do 2 rewrite opensAT_str.
                  rewrite map_shifts_bnds.
                  rewrite <- Len2. rewrite <- freeTs_length.
                  rewrite <- substsATR_opensATR.
                  rewrite <- map_shifts_bnds. rewrite <- opensATR_concat.
                  rewrite opensATR_transR. 2: apply lclosedTs_vars.
                  rewrite opensATR_transR. 2: apply lclosedTs_concat; apply lclosedTs_vars.
                  rewrite <- freeTs_concat.
                  assert (HG1' : okG (C & as1~~ks1 & as2~~ks2) (transG G & x1 ~ strT (transR (opensAR (freeTs as1) r1)))).
                    eapply weakeningDG_head. apply HG1. apply HC2.
                  assert (HG2 : okG (C & as1~~ks1 & as2~~ks2) (transG G & x1 ~ strT (transR (opensAR (freeTs as1) r1)) & x2 ~ strT (transR (opensAR (freeTs (as1 ++ as2)) r2)))).
                    apply okG_var.
                      apply HG1'. fresh_solve.
                      apply ofT_str. apply IHr2.
                  rewrite Hr1'.
                    2: rewrite <- Len1; rewrite <- Len2; apply fresh_concat_l; try solve [rewrite Len1 | rewrite Len2]; fresh_solve.
                    2: omega.
                  apply ofER_cat.
                    apply ofER_dots'.
                      rewrite <- opensATR_transR. 2: apply lclosedTs_vars. rewrite labsR_opensATR.
                        rewrite transR_diff. apply distinct_diffR. rewrite <- labsR_transR. apply Hr1.
                      rewrite opensAR_diff. rewrite transR_diff. intros l Elem. eapply proj1. apply getR_diff. apply Elem.
                      apply ofE_var.
                        apply HG2.
                        apply binds_mid. eapply regularityG. apply HG2.
                    apply ofER_dots.
                      rewrite <- opensATR_transR. 2: apply lclosedTs_vars. rewrite labsR_opensATR. rewrite <- labsR_transR. apply Hr2.
                      apply ofE_var.
                        apply HG2.
                        apply binds_head.
                  intros i Elem. eapply bnd_upper. apply Elem.
                  apply lclosedTs_vars.
                  apply lclosedTs_vars.
            eapply weakeningDT_head. eapply weakeningDT_head. rewrite exsT_concat. apply Ht. apply HC1. apply HC2.
          eapply weakeningDT_head. apply Ht. apply HC1.
        apply Ht.
    Case "InclB".
      intros C G M q e HM IHe.
      apply IHe.

    Case "ModP".
      intros C G M s ks e HM IHe Hs.
      simpl transQ in *.
      apply ofE_unpacks with (L := dom C \u dom (transG G)).
        apply IHe.
        intros as' x Fras' Frx.
          assert (HC' : okD (C & as'~~ks)).
            apply okD_vars. eapply validityDE. apply IHe. fresh_solve.
          rewrite openXE_opensAE. 2: apply lclosedE_var. unfold openXE. simpl substXE.
          rewrite opensAE_lclosed. 2: apply lclosedE_var.
          rewrite opensAT_lclosed. 2: eapply regularityT; apply Hs.
          apply ofE_var.
            apply okG_var.
              apply weakeningDG_head. eapply validityGE. apply IHe. apply HC'.
              notin_solve.
            apply weakeningDT_head. apply Hs. apply HC'.
            apply binds_head.
        apply Hs.
Qed.

Corollary sound_elabT : forall C G T k t,
  C; G |=T T ~: k ~> t ->
  ofT C t k.
Proof.
  intros C G T k t HT.
  eapply proj2.
  generalize C G T k t HT.
  eapply proj1. apply sound_elab.
Qed.

Corollary sound_elabE : forall C G E t e,
  C; G |=E E ~: t ~> e ->
  ofE C (transG G) e t.
Proof.
  eapply proj1. eapply proj2. apply sound_elab.
Qed.

Corollary sound_elabS : forall C G S q,
  C; G |=S S ~> q ->
  ofT C (transQ q) omK.
Proof.
  intros C G S q HS.
  eapply proj2.
  generalize C G S q HS.
  eapply proj1. do 2 eapply proj2. apply sound_elab.
Qed.

Corollary sound_elabD : forall C G D q,
  C; G |=D D ~> q ->
  ofT C (transQ q) omK.
Proof.
  intros C G D q HD.
  eapply proj2.
  generalize C G D q HD.
  eapply proj1. do 3 eapply proj2. apply sound_elab.
Qed.

Corollary sound_elabM : forall C G M q e,
  C; G |=M M ~: q ~> e ->
  ofE C (transG G) e (transQ q).
Proof.
  eapply proj1. do 4 eapply proj2. apply sound_elab.
Qed.

Corollary sound_elabB : forall C G B q e,
  C; G |=B B ~: q ~> e ->
  ofE C (transG G) e (transQ q).
Proof.
  eapply proj1. do 5 eapply proj2. apply sound_elab.
Qed.

Corollary sound_elabP : forall C G P s e,
  C; G |=P P ~: s ~> e ->
  ofE C (transG G) e (transS s).
Proof.
  do 6 eapply proj2. apply sound_elab.
Qed.

